﻿using System;
using System.Collections.Generic;
using System.IO;

namespace Jacquot_Timepix_V2.Models
{
    public class Data
    {
        private static readonly char[] _wordSeparators = new char[] { '\t', ' ' };

        private static bool? _usingDotSeparator = null;
        private static bool UsingDotSeparator
        {
            get
            {
                if (_usingDotSeparator == null)
                {
                    // test if this machine uses dot or comma separators
                    double testNum = 1.1;
                    _usingDotSeparator = double.TryParse("1.1", out double parsed) && (parsed == testNum);
                }
                return (bool)_usingDotSeparator;
            }
        }

        public string FilePath { get; }
        public string FileName { get; }
        public List<Cluster> Clusters { get; } = new();

        /// <summary>
        /// Use <see cref="Load"/> method to generate Data objects
        /// </summary>
        private Data(string filepath)
        {
            FilePath = filepath;
            FileName = Path.GetFileName(filepath);
        }

        public static Data Load(string filepath)
        {
            Data data = new(filepath);

            uint id = 0;
            ushort x = 0, y = 0;
            uint time = 0;
            ushort size = 0;
            double energy = 0;
            bool clusterOK = false;

            List<Pixel> pixels = new();

            // parse everything
            foreach (string line in System.IO.File.ReadLines(filepath))
            {
                if (!String.IsNullOrEmpty(line))
                {
                    string[] words = UsingDotSeparator ? line.Split(_wordSeparators) : line.Replace('.', ',').Split(_wordSeparators);

                    switch (line[0])
                    {
                        case '-':
                            // Cluster information

                            pixels = new();
                            clusterOK = true;
                            clusterOK &= uint.TryParse(words[2], out id);
                            clusterOK &= ushort.TryParse(words[4], out x);
                            clusterOK &= ushort.TryParse(words[6], out y);
                            clusterOK &= uint.TryParse(words[8], out time);
                            clusterOK &= ushort.TryParse(words[10], out size);
                            clusterOK &= double.TryParse(words[12], out energy);

                            break;
                        case 'X':
                            // Pixel information

                            if (!clusterOK)
                            {
                                // don't read more cluster data if previous data was invalid
                                break;
                            };

                            if (ushort.TryParse(words[1], out ushort pX) && ushort.TryParse(words[3], out ushort pY) && ushort.TryParse(words[5], out ushort tot))
                            {
                                pixels.Add(new(pX, pY, tot));
                            }

                                break;
                        case 'C':
                            // Cluster type info

                            if (!clusterOK)
                            {
                                // don't read more cluster data if previous data was invalid
                                break;
                            }

                            // all cluster information is prepared, create Cluster object
                            if (clusterOK && TryParseTypes(words, out ClusterType? physical, out RadiationType? particle) && physical != null && particle != null)
                            {
                                Cluster? cluster = Cluster.Create(id, x, y, time, size, energy, (ClusterType)physical, (RadiationType)particle, pixels);
                                if (cluster != null)
                                {
                                    data.Clusters.Add(cluster);
                                }
                            }

                            break;
                        default:
                            break;
                    }
                }
            }
            return data;
        }

        /// <returns><c>true</c> if both enums were parsed correctly</returns>
        private static bool TryParseTypes(string[] words, out ClusterType? physical, out RadiationType? particle)
        {
            // remove ',' last char
            string physicalCandidate = words[3][..^1];
            string particleCandidate = words[5];
#if DEBUG
            string error = "";
            if (!Enum.IsDefined(typeof(ClusterType), physicalCandidate))
            {
                error += "Physical \"" + physicalCandidate + "\" is undefined!\n";
            }
            if (!Enum.IsDefined(typeof(RadiationType), particleCandidate))
            {
                error += "Particle \"" + particleCandidate + "\" is undefined!\n";
            }
            if (!String.IsNullOrEmpty(error))
            {
                System.Windows.MessageBox.Show(error);
            }
#endif
            object? result;
            bool success = true;
            if (Enum.TryParse(typeof(ClusterType), physicalCandidate, out result))
            {
                physical = (ClusterType?)result;
            }
            else
            {
                physical = null;
                success &= false;
            }

            if(Enum.TryParse(typeof(RadiationType), particleCandidate, out result))
            {
                particle = (RadiationType?)result;
            }
            else
            {
                particle = null;
                success &= false;
            }

            return success;
        }

#region override ==
        public override bool Equals(object? obj)
        {
            if (obj is not Data data)
                return false;

            return FileName == data.FileName;
        }

        public override int GetHashCode()
        {
            return FileName.GetHashCode();
        }

        public static bool operator ==(Data data1, Data data2)
        {
            if (data1 is null && data2 is null)
            {
                return true;
            }
            return data1 is not null && data1.Equals(data2);
        }

        public static bool operator !=(Data data1, Data data2)
        {
            return !(data1 == data2);
        }
#endregion
    }
}
