﻿using ScottPlot;
using ScottPlot.Plottable;
using System;
using System.Collections.Generic;
using System.Linq;

namespace Jacquot_Timepix_V2.Models
{
    public class File
    {
        public Data Data { get; }
        public int ClusterCount => Data.Clusters.Count;

        public bool _show = true;
        public bool Show
        {
            get => _show;
            set
            {
                if (_show != value)
                {
                    _show = value;
                    if (value)
                    {
                        _histograms.ForEach(h => h.PlotView.Plot.Add(h.Plottable));
                    }
                    else
                    {
                        _histograms.ForEach(h => h.PlotView.Plot.Remove(h.Plottable));
                    }
                    Refresh();
                }
            }
        }

        /// <summary>
        /// A list of histogram views and their histogram plots
        /// </summary>
        private readonly List<(WpfPlot PlotView, SignalPlotXYConst<double, double> Plottable)> _histograms = new();

        public File(string filepath)
        {
#if DEBUG
            string pad = "";
            string fileMsg = "█ Loading file: " + filepath + " █";
            foreach (char _ in fileMsg)
            {
                pad += '█';
            }
            Console.WriteLine('\n' + pad);
            Console.WriteLine(fileMsg);
            Console.WriteLine(pad);

            //_ = Console.ReadKey();
#endif
            Data = Data.Load(filepath);
        }

        public void InitializeHistograms(List<WpfPlot> histograms)
        {
            // Histogram FillRatio
            InitializeHistogram(histograms[0], Data.Clusters.Select(c => c.FillRatio).Where(f => f > 0), 0.04);

            // Histogram Linearity
            InitializeHistogram(histograms[1], Data.Clusters.FindAll(c => c.Linearity > 0 && c.Size > 2).Select(c => c.Linearity), 0.05);

            // Histogram Size
            InitializeHistogram(histograms[2], Data.Clusters.Select(c => c.Size));

            // Histogram ToT
            InitializeHistogram(histograms[3], Data.Clusters.SelectMany(c => c.Pixels).Select(p => p.Tot), removeLast: 5);

            // Histogram Variation
            InitializeHistogram(histograms[4], Data.Clusters.Select(c => c.StandardDeviation / c.AverageTot).Where(v => v > 0), 0.05);
            
            Refresh();
        }

        private void InitializeHistogram<TX>(WpfPlot target, IEnumerable<TX> values, double binSize = 0, int removeLast = 0)
            where TX : unmanaged, IComparable, IConvertible, IEquatable<TX>
        {
            double[] xs, ys;

            double[] dValues = GenericToDouble(values);

            if (binSize == 0)
            {
                (xs, ys) = Histogram(dValues);
            }
            else
            {
                (ys, xs) = ScottPlot.Statistics.Common.Histogram(dValues, dValues.Min(), dValues.Max(), binSize);
            }

            // ToT pixels can be completely saturated causing an extremely tall peak at the end of the graph for max ToT
            if (removeLast > 0)
            {
                for (int i = 1; i <= removeLast; i++)
                {
                    ys[^i] = 0;
                }
            }

            // normalize
            double m = 100 / ys.Max();
            for (int i = 0; i < ys.Length; i++)
            {
                ys[i] *= m;
            }

            _histograms.Add((target, target.Plot.AddSignalXYConst(xs, ys, label: Data.FileName)));
        }

        /// <summary>
        /// Converts a list of values into a histogram
        /// </summary>
        /// <typeparam name="TX">Generic type that must be convertible to a double</typeparam>
        /// <returns>Ordered histogram of x/y points in 2 arrays</returns>
        public static (double[] xs, double[] ys) Histogram<TX>(IEnumerable<TX> values)
            where TX : unmanaged, IComparable, IConvertible, IEquatable<TX>
        {
            if (values == null)
            {
                throw new ArgumentNullException();
            }

            // Dictionary<tot value, how many tots of this value>
            Dictionary<TX, uint> histogram = new();
            foreach (TX value in values)
            {
                if (!histogram.TryAdd(value, 1))
                {
                    // increment value if entry already exists
                    histogram[value]++;
                }
            }

            double[] xs = new double[histogram.Count];
            double[] ys = new double[histogram.Count];

            // convert dictionary to 2 arrays
            int i = 0;
            foreach (KeyValuePair<TX, uint> point in histogram.OrderBy(x => x.Key))
            {
                xs[i] = Convert.ToDouble(point.Key);
                ys[i++] = point.Value;
            }

            return (xs, ys);
        }

        private static double[] GenericToDouble<T>(IEnumerable<T> values)
            where T : unmanaged, IComparable, IConvertible, IEquatable<T>
        {
            double[] ans = new double[values.Count()];
            int i = 0;
            foreach (T value in values)
            {
                ans[i++] = Convert.ToDouble(value);
            }
            return ans;
        }
        private void Refresh()
        {
            _histograms.ForEach(h => h.PlotView.Refresh());
        }

#region override ==
        public override bool Equals(object? obj)
        {
            if (obj is not File file)
                return false;

            return Data == file.Data;
        }

        public override int GetHashCode()
        {
            return Data.GetHashCode();
        }

        public static bool operator ==(File file1, File file2)
        {
            if (file1 is null && file2 is null)
            {
                return true;
            }
            return file1 is not null && file1.Equals(file2);
        }

        public static bool operator !=(File file1, File file2)
        {
            return !(file1 == file2);
        }
#endregion
    }
}
